import json
import os

import pandas as pd
import requests
import websocket
from IPython.core.magic import Magics, line_cell_magic, magics_class, needs_local_scope

from thanosql.exception import (
    ThanoSQLConnectionError,
    ThanoSQLInternalError,
    ThanoSQLSyntaxError,
)
from thanosql.parse import *
from thanosql.util import format_result

engine_cluster_ip = os.getenv("THANOSQL_ENGINE_SERVICE_HOST")
DEFAULT_API_URL = f"ws://{engine_cluster_ip}/ws/v1/query"


def request_thanosql_engine(ws, api_url, api_token, query_context):
    try:
        ws.connect(f"{api_url}?api_token={api_token}")
    except:
        raise ThanoSQLConnectionError("Could not connect to the Websocket")

    ws.send(json.dumps(query_context))

    try:
        while True:
            output = ws.recv()
            output_dict = json.loads(output)

            if output_dict["output_type"] == "MESSAGE":
                print(output_dict["output_message"])
            elif output_dict["output_type"] == "RESULT":
                return format_result(output_dict["output_message"])
            elif output_dict["output_type"] == "ERROR":
                raise ThanoSQLInternalError(output_dict["output_message"])

    except KeyboardInterrupt:
        ws.close()
        raise ThanoSQLConnectionError("KeyboardInterrupt")


@magics_class
class ThanosMagic(Magics):
    ws = websocket.WebSocket()

    @needs_local_scope
    @line_cell_magic
    def thanosql(self, line: str = None, cell: str = None, local_ns={}):
        if line:
            if is_url(line):
                # Set API URL
                api_url = line.strip()
                os.environ["API_URL"] = api_url
                print(f"API URL is changed to {api_url}")
                return

            elif is_api_token(line):
                # Set API Token
                api_token = line.strip().split("API_TOKEN=")[-1]
                os.environ["API_TOKEN"] = api_token
                print(f"API Token is set as '{api_token}'")
                return

            # 'line' will be treated the same as 'cell'
            else:
                cell = line

        if not cell:
            return

        api_url = os.getenv("API_URL", DEFAULT_API_URL)
        api_token = os.getenv("API_TOKEN", None)

        if not api_token:
            raise ThanoSQLConnectionError(
                "An API Token is required. Set the API Token by running the following: %thanosql API_TOKEN=<API_TOKEN>"
            )

        query_string = cell

        if is_multiple_queries(query_string):
            raise ThanoSQLSyntaxError("Multiple Queries are not supported.")

        query_context = {"query_string": query_string, "query_type": "thanosql"}

        return request_thanosql_engine(self.ws, api_url, api_token, query_context)

    @needs_local_scope
    @line_cell_magic
    def psql(self, line: str = None, cell: str = None, local_ns={}):
        if line:
            cell = line

        if not cell:
            return

        api_url = os.getenv("API_URL", DEFAULT_API_URL)
        api_token = os.getenv("API_TOKEN", None)

        query_string = cell

        query_context = {"query_string": query_string, "query_type": "psql"}

        return request_thanosql_engine(self.ws, api_url, api_token, query_context)


# In order to actually use these magics, you must register them with a
# running IPython.
def load_ipython_extension(ipython):
    """Load the extension in IPython."""
    ipython.register_magics(ThanosMagic)



# ThanoSQL Error Handling
class ThanoSQLConnectionError(Exception):
    def __init__(self, msg):
        super().__init__(msg)


class ThanoSQLSyntaxError(Exception):
    def __init__(self, msg):
        super().__init__(msg)


class ThanoSQLInternalError(Exception):
    def __init__(self, msg):
        super().__init__(msg)



import pandas as pd
from IPython.display import Audio, Image, Video, display
from sqlalchemy import create_engine, text
from sqlalchemy.exc import ResourceClosedError

from thanosql.exception import ThanoSQLConnectionError, ThanoSQLInternalError


def format_result(output_dict: dict):

    data = output_dict["data"]
    workspace_db_info = data.get("workspace_db_info")
    response_type = data.get("response_type")

    query_string = text(data.get("query_string"))
    extra_query_string = text(data.get("extra_query_string"))

    user = workspace_db_info.get("user")
    password = workspace_db_info.get("password")
    database = workspace_db_info.get("database")
    host = workspace_db_info.get("host")

    connection_string = f"postgresql://{user}:{password}@/{database}?host={host}"

    try:
        engine = create_engine(connection_string)
    except:
        raise ThanoSQLConnectionError("Error connecting to workspace database")

    with engine.connect() as conn:
        result = None

        if response_type == "NORMAL":
            try:
                result = pd.read_sql_query(query_string, conn)
            except ResourceClosedError:
                """
                ResourceClosedError will capture queries
                like INSERT and DROP that don’t return a value.
                This is not the best solution as we are presumptuously assuming
                that the connection with the database will always be secure and succeed.
                If a failure happens in the database,
                ResourceClosedError will be raised
                and “Success” will be printed out, which is a problem.
                Therefore, this is subject to change in the future.
                """
                print("Success")

        elif response_type == "SELECT":
            result = pd.read_sql_query(query_string, conn)

        elif response_type == "SELECT_DROP":
            result = pd.read_sql_query(query_string, conn)
            conn.execute(extra_query_string)

        elif response_type is None:
            print("Success")

    # close sqlalchemy(DB) engine
    engine.dispose()

    print_type = data.get("print")
    if print_type:
        print_option = data.get("print_option", {})
        return print_result(result, print_type, print_option)

    return result


def print_result(query_df, print_type: str, print_option):
    if print_type == "print_image":
        return print_image(query_df, print_option)
    elif print_type == "print_audio":
        return print_audio(query_df, print_option)
    elif print_type == "print_video":
        return print_video(query_df, print_option)
    else:
        raise ThanoSQLInternalError("Error: Wrong print_type.")


def print_image(df, print_option):
    column_name = print_option.get("image_col", "image_path")
    image_file_list = list(df[column_name])

    base_dir = print_option.get("base_dir", "")
    limit = print_option.get("limit")
    for image_path in image_file_list[:limit]:
        image_full_path = f"{base_dir}/{image_path}"
        print(image_full_path)
        display(Image(image_full_path, width=240, height=240))
    return


def print_audio(df, print_option):
    column_name = print_option.get("audio_col", "audio_path")
    audio_file_list = list(df[column_name])

    base_dir = print_option.get("base_dir", "")
    limit = print_option.get("limit")
    for audio_path in audio_file_list[:limit]:
        audio_full_path = f"{base_dir}/{audio_path}"
        print(audio_full_path)
        display(Audio(audio_full_path))
    return


def print_video(df, print_option):
    column_name = print_option.get("video_col", "video_path")
    video_file_list = list(df[column_name])

    base_dir = print_option.get("base_dir", "")
    limit = print_option.get("limit")
    for video_path in video_file_list[:limit]:
        video_full_path = f"{base_dir}/{video_path}"
        print(video_full_path)
        display(Video(video_full_path, embed=True))
    return
